/*
 * Memory allocator wrappers
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

/* NOTES:
 *
 * Requested allocation size is checked agains SSIZE_MAX
 * to prevent integer overflow. This allows to catch
 * situations and terminate program if the requested size
 * has negative value.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include "macros.h"
#include "common.h"
#include "failures.h"

void *xmalloc(size_t size)
{
	void *ptr;

	if (size == 0 || size > SSIZE_MAX)
		xerror(1, "xmalloc: requested %lu bytes", (unsigned long) size);

	ptr = malloc(size);

	if (ptr == NULL) {
		xerror(1, "xmalloc: out of memory allocating %lu bytes",
		      (unsigned long) size);
	}

	return ptr;
}

void *xmalloc0(size_t size)
{
	void *ptr;

	if (size == 0 || size > SSIZE_MAX)
		xerror(1, "xmalloc0: requested %lu bytes", (unsigned long) size);

	ptr = malloc(size);

	if (ptr == NULL) {
		xerror(1, "xmalloc0: out of memory allocating %lu bytes",
		      (unsigned long) size);
	}

	memset(ptr, 0, size);

	return ptr;
}

void *xcalloc(size_t nmemb, size_t size)
{
	void *ptr;

	if (nmemb == 0 || nmemb > SSIZE_MAX)
		xerror(1, "xcalloc requested %lu members", (unsigned long) nmemb);

	if (size == 0 || size > SSIZE_MAX)
		xerror(1, "xcalloc requested %lu bytes", (unsigned long) size);

	ptr = calloc(nmemb, size);

	if (ptr == NULL) {
		xerror(1, "xcalloc out of memory allocating %lu bytes",	(unsigned long) size);
	}

	return ptr;
}

void *xrealloc(void *mem, size_t new_size)
{
	void *ptr;

	if (new_size == 0 || new_size > SSIZE_MAX)
		xerror(1, "xrealloc: requested %lu bytes", (unsigned long) new_size);

	if (mem == NULL)
		ptr = malloc(new_size);
	else
		ptr = realloc(mem, new_size);

	if (ptr == NULL) {
		xerror(1, "xrealloc: out of memory allocating %lu bytes",
		      (unsigned long) new_size);
	}

	return ptr;
}

void *xrealloc0(void *mem, size_t old_size, size_t new_size)
{
	void *ptr;

	if (mem == NULL && old_size)
		xerror(1, "xrealloc0: old_size != 0 on NULL memory");

	if (new_size == 0 || new_size > SSIZE_MAX)
		xerror(1, "xrealloc0: requested %lu bytes",
		      (unsigned long) new_size);

	if (mem == NULL)
		ptr = malloc(new_size);
	else
		ptr = realloc(mem, new_size);

	if (ptr == NULL) {
		xerror(1, "xrealloc0 out: of memory allocating %lu bytes",
		      (unsigned long) new_size);
	}

	if (new_size > old_size)
		memset(ptr + old_size, 0, new_size - old_size);

	return ptr;
}

void xfree(void *ptr)
{
	if (ptr == NULL)
		xerror(1, "xfree: NULL pointer");

	free(ptr);
}

char *xstrdup(const char *str)
{
	char *res;
	size_t len;

	if (!str)
		xerror(1, "xstrdup: str NULL");

	len = strlen(str);
	res = xmalloc0(len + 1);
	memcpy(res, str, len);
	res[len] = '\0';

	return res;
}

char *xstrndup(const char *str, size_t max)
{
	char *res;
	size_t len;

	if (!str)
		xerror(1, "xstrndup: str NULL");

	len = strnlen(str, max);
	res = xmalloc0(len+1);
	memcpy(res, str, len);
	res[len] = '\0';

	return res;
}

char *xstrncpy(char *dst, const char *src, size_t max)
{
	size_t len;

	if (!dst || !src)
		xerror(1, "xstrncpy: dst or src NULL");

	if (max > SSIZE_MAX)
		xerror(1, "xstrncpy: max %lu bytes", (unsigned long) max);

	len = strnlen(src, max-1);
	memcpy(dst, src, len);
	dst[len] = '\0';

	return dst;
}

#define XGETCWD_BUFLEN 64

char *xgetcwd()
{
	size_t len = 0;
	char *buf = NULL, *ret;

	do {
		/* Allocate some space in buffer. */
		len += XGETCWD_BUFLEN;
		buf = xrealloc(buf, len);
		/* Write current directory to buffer. */
		ret = getcwd(buf, len);
		/* Retry with increased buffer if failed. */
	} while (ret == NULL && errno == ERANGE);

	if (ret == NULL)
		xerror(1, "getcwd: %s", strerror(errno));

	return buf;
}
